@@ -13,6 +13,10 @@ from .dataset import (
     WebDataset,
     WebLoader,
 )
+from .pipeline import (
+    DataPipeline,
+    stage,
+)
 from .composable import (
     Composable,
     Shorthands,
@@ -22,10 +26,17 @@ from .shardlists import (
     SimpleShardList,
     PytorchShardList,
     ResampledShards,
+    split_by_node,
+    split_by_worker,
+    non_empty,
+    resampled,
 )
 from .extradatasets import (
     MockDataset
 )
+from .tariterators import (
+    tarfile_samples,
+)
 from .handlers import (
     reraise_exception,
     ignore_and_continue,
@@ -50,9 +61,12 @@ from .tariterators import (
     tar_file_expander,
     group_by_keys,
 )
-
 from .iterators import (
     transform_with,
+    getfirst,
+)
+from .filters import (
+    Curried,
     info,
     shuffle,
     select,
@@ -65,6 +79,6 @@ from .iterators import (
     map_tuple,
     batched,
     unbatched,
-    getfirst,
 )
 from .dbcache import DBCache
+from .dsspecs import RoundRobin
