@@ -11,22 +11,25 @@ Code works locally or over HTTP connections.
 """
 
 import os
-import sys
 import random
-import yaml
-from dataclasses import dataclass
+import sys
+import time
+from dataclasses import dataclass, field
+from itertools import islice
 from typing import List
 
 import braceexpand
+import yaml
 
-from .pytorch import IterableDataset
 from .composable import Composable
+from .filters import Curried
+from .pytorch import IterableDataset
 
 
 class SimpleShardList(IterableDataset, Composable):
     """An iterable dataset yielding a list of urls."""
 
-    def __init__(self, urls):
+    def __init__(self, urls, seed=None):
         """Iterate through the list of shards.
 
         :param urls: a list of URLs as a Python list or brace notation string
@@ -38,13 +41,75 @@ class SimpleShardList(IterableDataset, Composable):
             urls = list(urls)
         self.urls = urls
         assert isinstance(self.urls[0], str)
+        self.seed = seed
+
+    def __len__(self):
+        return len(self.urls)
 
     def __iter__(self):
         """Return an iterator over the shards."""
-        for url in self.urls:
+        urls = self.urls.copy()
+        if self.seed is not None:
+            random.Random(self.seed).shuffle(urls)
+        for url in urls:
             yield dict(url=url)
 
 
+def split_by_node(src, group=None):
+    import torch.distributed
+
+    if torch.distributed.is_available() and torch.distributed.is_initialized():
+        group = group or torch.distributed.group.WORLD
+        rank = torch.distributed.get_rank(group=group)
+        size = torch.distributed.get_world_size(group=group)
+        for s in islice(src, rank, None, size):
+            yield s
+    else:
+        for s in src:
+            yield s
+
+
+def split_by_worker(src):
+    import torch.utils.data
+
+    winfo = torch.utils.data.get_worker_info()
+    if winfo is None:
+        for s in src:
+            yield s
+    else:
+        for s in islice(src, winfo.id, None, winfo.num_workers):
+            yield s
+
+
+def resampled_(src, n=sys.maxsize):
+    import random
+
+    seed = time.time_ns()
+    try:
+        seed = open("/dev/random", "rb").read(20)
+    except Exception as exn:
+        print(repr(exn)[:50], file=sys.stderr)
+    rng = random.Random(seed)
+    items = list(src)
+    for i in range(n):
+        yield rng.choice(items)
+
+
+resampled = Curried(resampled_)
+
+
+def non_empty(src):
+    count = 0
+    for s in src:
+        yield s
+        count += 1
+    if count == 0:
+        raise ValueError("pipeline stage received no data at all and this was declared as an error")
+
+
+# BELOW IS DEPRECATED ###
+
+
 class PytorchEnv:
     """A class encapsulating the PyTorch node/worker environment."""
 
@@ -116,7 +181,7 @@ class MSSource:
     name: str = ""
     perepoch: int = -1
     resample: bool = False
-    urls: List[str] = None
+    urls: List[str] = field(default_factory=list)
 
 
 default_rng = random.Random()
@@ -268,6 +333,7 @@ class ResampledShards(IterableDataset, Composable):
         self,
         urls,
         nshards=sys.maxsize,
+        env=None,
     ):
         """Sample shards from the shard list with replacement.
 
@@ -280,9 +346,15 @@ class ResampledShards(IterableDataset, Composable):
             urls = list(urls)
         self.urls = urls
         self.nshards = nshards
+        self.env = env or PytorchEnv()
+        self.rng = random.Random()
         assert isinstance(self.urls[0], str)
 
     def __iter__(self):
         """Return an iterator over the shards."""
+        rank, world = self.env.rank or (0, 1)
+        worker, nworkers = self.env.worker or (0, 1)
+        seed = (rank, worker, os.getpid(), time.time())
+        self.rng.seed(seed)
         for _ in range(self.nshards):
-            yield dict(url=random.choice(self.urls))
+            yield dict(url=self.rng.choice(self.urls))
