@@ -15,6 +15,7 @@ import braceexpand
 from .handlers import reraise_exception
 
 from . import gopen
+from . import filters
 
 
 trace = False
@@ -106,6 +107,7 @@ def tar_file_iterator(fileobj, skip_meta=r"__[^/]*__($|/)", handler=reraise_exce
             result = dict(fname=fname, data=data)
             result.update(info)
             yield result
+            stream.members = []
         except Exception as exn:
             if hasattr(exn, "args") and len(exn.args) > 0:
                 exn.args = (exn.args[0] + " @ " + str(fileobj),) + exn.args[1:]
@@ -171,3 +173,12 @@ def group_by_keys(data, keys=base_plus_ext, lcase=True, suffixes=None, handler=N
             current_sample[suffix] = value
     if valid_sample(current_sample):
         yield current_sample
+
+def tarfile_samples(src, handler=reraise_exception):
+    streams = url_opener(src, handler=handler)
+    files = tar_file_expander(streams, handler=handler)
+    samples = group_by_keys(files, handler=handler)
+    return samples
+
+
+tarfile_sampler = filters.Curried(tarfile_samples)
