@@ -1,10 +1,10 @@
 #
-# Copyright (c) 2017-2019 NVIDIA CORPORATION. All rights reserved.
+# Copyright (c) 2017-2021 NVIDIA CORPORATION. All rights reserved.
 # This file is part of the WebDataset library.
 # See the LICENSE file for licensing terms (BSD-style).
 #
 
-__all__ = "TarWriter ShardWriter".split()
+"""Classes and functions for writing tar files and WebDataset files."""
 
 import io
 import pickle
@@ -12,14 +12,14 @@ import re
 import tarfile
 import time
 import json
+from typing import Union, Any, Callable, Optional
 
 import numpy as np
-import PIL
 
 from . import gopen
 
 
-def imageencoder(image, format="PNG"):  # skipcq: PYL-W0622
+def imageencoder(image: Any, format: str = "PNG"):  # skipcq: PYL-W0622
     """Compress an image using PIL and return it as a string.
 
     Can handle float or uint8 images.
@@ -28,14 +28,19 @@ def imageencoder(image, format="PNG"):  # skipcq: PYL-W0622
     :param format: compression format (PNG, JPEG, PPM)
 
     """
+    import PIL
+
+    assert isinstance(image, (PIL.Image.Image, np.ndarray)), type(image)
+
     if isinstance(image, np.ndarray):
         if image.dtype in [np.dtype("f"), np.dtype("d")]:
             if not (np.amin(image) > -0.001 and np.amax(image) < 1.001):
-                raise ValueError(
-                    f"image values out of range {np.amin(image)} {np.amax(image)}"
-                )
+                raise ValueError(f"image values out of range {np.amin(image)} {np.amax(image)}")
             image = np.clip(image, 0.0, 1.0)
             image = np.array(image * 255.0, "uint8")
+        assert image.ndim in [2, 3]
+        if image.ndim == 3:
+            assert image.shape[2] in [1, 3]
         image = PIL.Image.fromarray(image)
     if format.upper() == "JPG":
         format = "JPEG"
@@ -50,7 +55,13 @@ def imageencoder(image, format="PNG"):  # skipcq: PYL-W0622
         return result.getvalue()
 
 
-def bytestr(data):
+def bytestr(data: Any):
+    """Convert data into a bytestring.
+
+    Uses str and ASCII encoding for data that isn't already in string format.
+
+    :param data: data
+    """
     if isinstance(data, bytes):
         return data
     if isinstance(data, str):
@@ -58,7 +69,13 @@ def bytestr(data):
     return str(data).encode("ascii")
 
 
-def torch_dumps(data):
+def torch_dumps(data: Any):
+    """Dump data into a bytestring using torch.dumps.
+
+    This delays importing torch until needed.
+
+    :param data: data to be dumped
+    """
     import io
     import torch
 
@@ -67,7 +84,33 @@ def torch_dumps(data):
     return stream.getvalue()
 
 
+def numpy_dumps(data: np.ndarray):
+    """Dump data into a bytestring using numpy npy format.
+
+    :param data: data to be dumped
+    """
+    import io
+    import numpy.lib.format
+
+    stream = io.BytesIO()
+    numpy.lib.format.write_array(stream, data)
+    return stream.getvalue()
+
+
+def numpy_npz_dumps(data: np.ndarray):
+    """Dump data into a bytestring using numpy npz format.
+
+    :param data: data to be dumped
+    """
+    import io
+
+    stream = io.BytesIO()
+    np.savez_compressed(stream, **data)
+    return stream.getvalue()
+
+
 def make_handlers():
+    """Create a list of handlers for encoding data."""
     handlers = {}
     for extension in ["cls", "cls2", "class", "count", "index", "inx", "id"]:
         handlers[extension] = lambda x: str(x).encode("ascii")
@@ -76,6 +119,10 @@ def make_handlers():
     for extension in ["png", "jpg", "jpeg", "img", "image", "pbm", "pgm", "ppm"]:
 
         def f(extension_):
+            """f.
+
+            :param extension_:
+            """
             handlers[extension] = lambda data: imageencoder(data, extension_)
 
         f(extension)
@@ -83,6 +130,10 @@ def make_handlers():
         handlers[extension] = pickle.dumps
     for extension in ["pth"]:
         handlers[extension] = torch_dumps
+    for extension in ["npy"]:
+        handlers[extension] = numpy_dumps
+    for extension in ["npz"]:
+        handlers[extension] = numpy_npz_dumps
     for extension in ["json", "jsn"]:
         handlers[extension] = lambda x: json.dumps(x).encode("utf-8")
     for extension in ["ten", "tb"]:
@@ -108,7 +159,13 @@ def make_handlers():
 default_handlers = {"default": make_handlers()}
 
 
-def encode_based_on_extension1(data, tname, handlers):
+def encode_based_on_extension1(data: Any, tname: str, handlers: dict):
+    """Encode data based on its extension and a dict of handlers.
+
+    :param data: data
+    :param tname: file extension
+    :param handlers: handlers
+    """
     if tname[0] == "_":
         if not isinstance(data, str):
             raise ValueError("the values of metadata must be of string type")
@@ -124,16 +181,24 @@ def encode_based_on_extension1(data, tname, handlers):
     return handler(data)
 
 
-def encode_based_on_extension(sample, handlers):
-    return {
-        k: encode_based_on_extension1(v, k, handlers) for k, v in list(sample.items())
-    }
+def encode_based_on_extension(sample: dict, handlers: dict):
+    """Encode an entire sample with a collection of handlers.
+
+    :param sample: data sample (a dict)
+    :param handlers: handlers for encoding
+    """
+    return {k: encode_based_on_extension1(v, k, handlers) for k, v in list(sample.items())}
 
 
-def make_encoder(spec):
+def make_encoder(spec: Union[bool, str, dict, Callable]):
+    """Make an encoder function from a specification.
+
+    :param spec: specification
+    """
     if spec is False or spec is None:
 
         def encoder(x):
+            """Do not encode at all."""
             return x
 
     elif callable(spec):
@@ -141,6 +206,7 @@ def make_encoder(spec):
     elif isinstance(spec, dict):
 
         def encoder(sample):
+            """Encode based on extension."""
             return encode_based_on_extension(sample, spec)
 
     elif isinstance(spec, str) or spec is True:
@@ -151,6 +217,7 @@ def make_encoder(spec):
             raise ValueError(f"no handler found for {spec}")
 
         def encoder(sample):
+            """Encode based on extension."""
             return encode_based_on_extension(sample, handlers)
 
     else:
@@ -187,13 +254,23 @@ class TarWriter:
     def __init__(
         self,
         fileobj,
-        user="bigdata",
-        group="bigdata",
-        mode=0o0444,
-        compress=None,
-        encoder=True,
-        keep_meta=False,
+        user: str = "bigdata",
+        group: str = "bigdata",
+        mode: int = 0o0444,
+        compress: Optional[bool] = None,
+        encoder: Union[None, bool, Callable] = True,
+        keep_meta: bool = False,
     ):
+        """Create a tar writer.
+
+        :param fileobj: stream to write data to
+        :param user: user for tar files
+        :param group: group for tar files
+        :param mode: mode for tar files
+        :param compress: desired compression
+        :param encoder: encoder function
+        :param keep_meta: keep metadata (entries starting with "_")
+        """
         if isinstance(fileobj, str):
             if compress is False:
                 tarmode = "w|"
@@ -217,9 +294,11 @@ class TarWriter:
         self.compress = compress
 
     def __enter__(self):
+        """Enter context."""
         return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
+        """Exit context."""
         self.close()
 
     def close(self):
@@ -229,16 +308,6 @@ class TarWriter:
             self.own_fileobj.close()
             self.own_fileobj = None
 
-    def dwrite(self, key, **kw):
-        """Convenience function for `write`.
-
-        Takes key as the first argument and key-value pairs for the rest.
-        Replaces "_" with ".".
-        """
-        obj = dict(__key__=key)
-        obj.update({k.replace("_", "."): v for k, v in kw.items()})
-        self.write(obj)
-
     def write(self, obj):
         """Write a dictionary to the tar file.
 
@@ -254,9 +323,7 @@ class TarWriter:
             if k[0] == "_":
                 continue
             if not isinstance(v, (bytes, bytearray, memoryview)):
-                raise ValueError(
-                    f"{k} doesn't map to a bytes after encoding ({type(v)})"
-                )
+                raise ValueError(f"{k} doesn't map to a bytes after encoding ({type(v)})")
         key = obj["__key__"]
         for k in sorted(obj.keys()):
             if k == "__key__":
@@ -282,18 +349,24 @@ class TarWriter:
 
 
 class ShardWriter:
-    """Like TarWriter but splits into multiple shards.
-
-    :param pattern: output file pattern
-    :param maxcount: maximum number of records per shard (Default value = 100000)
-    :param maxsize: maximum size of each shard (Default value = 3e9)
-    :param kw: other options passed to TarWriter
-
-    """
+    """Like TarWriter but splits into multiple shards."""
 
     def __init__(
-        self, pattern, maxcount=100000, maxsize=3e9, post=None, start_shard=0, **kw
+        self,
+        pattern: str,
+        maxcount: int = 100000,
+        maxsize: float = 3e9,
+        post: Optional[Callable] = None,
+        start_shard: int = 0,
+        **kw,
     ):
+        """Create a ShardWriter.
+
+        :param pattern: output file pattern
+        :param maxcount: maximum number of records per shard (Default value = 100000)
+        :param maxsize: maximum size of each shard (Default value = 3e9)
+        :param kw: other options passed to TarWriter
+        """
         self.verbose = 1
         self.kw = kw
         self.maxcount = maxcount
@@ -310,6 +383,7 @@ class ShardWriter:
         self.next_stream()
 
     def next_stream(self):
+        """Close the current stream and move to the next."""
         self.finish()
         self.fname = self.pattern % self.shard
         if self.verbose:
@@ -327,11 +401,11 @@ class ShardWriter:
         self.size = 0
 
     def write(self, obj):
-        if (
-            self.tarstream is None
-            or self.count >= self.maxcount
-            or self.size >= self.maxsize
-        ):
+        """Write a sample.
+
+        :param obj: sample to be written
+        """
+        if self.tarstream is None or self.count >= self.maxcount or self.size >= self.maxsize:
             self.next_stream()
         size = self.tarstream.write(obj)
         self.count += 1
@@ -339,6 +413,7 @@ class ShardWriter:
         self.size += size
 
     def finish(self):
+        """Finish all writing (use close instead)."""
         if self.tarstream is not None:
             self.tarstream.close()
             assert self.fname is not None
@@ -347,6 +422,7 @@ class ShardWriter:
             self.tarstream = None
 
     def close(self):
+        """Close the stream."""
         self.finish()
         del self.tarstream
         del self.shard
@@ -354,7 +430,9 @@ class ShardWriter:
         del self.size
 
     def __enter__(self):
+        """Enter context."""
         return self
 
     def __exit__(self, *args, **kw):
+        """Exit context."""
         self.close()
